// Copyright (C) 1999-2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of ccscript.
// 
// The exception is that, if you link the ccscript library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the ccscript library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name ccscript.  If you copy code from other releases into a copy of
// ccscript, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for ccscript, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#include <cc++/config.h>
#include <cc++/macros.h>
#include <cc++/misc.h>
#include <cc++/slog.h>
#include <cc++/url.h>
#include <cstdlib>
#include <cstdio>
#include "script.h"

using namespace std;

static void logerror(const char *script, unsigned id, const char *msg)
{
	slog(SLOG_ERROR) << script;
	if(id)
		slog() << "(" << id << ")";
	slog() << ": " << msg << endl;
}

static void adjustValue(char *buffer, const char *cp, int diff)
{
	int value = atoi(buffer);
	int len = strlen(buffer);
	time_t now;
	struct tm *dt, dtbuff;

	if(!cp)
		cp = "";

	if(!strnicmp(cp, "hour", 4))
	{
		if(len == 6)
		{
			value = atoi(buffer + 4);
			buffer[4] = 0;
			value += atoi(buffer + 2) * 60;
			buffer[2] = 0;
			value += atoi(buffer) * 3600;
			value += 3600 * diff;
			value %= (3600 * 24);
			sprintf(buffer, "%02d%02d%02",
				value / 3600,
				(value % 3600) / 60,
				value % 60);
			return;
		}
		else if(len == 8)
			return;
		now = atol(buffer);
		now += (diff * 3600);
		sprintf(buffer, "%10ld", now);
		return;
	}	
	else if(!strnicmp(cp, "min", 3))
	{
		if(len == 6)
		{
			value = atoi(buffer + 4);
			buffer[4] = 0;
			value += 60 * atoi(buffer + 2);
			buffer[2] = 0;
			value += 3600 * atoi(buffer);
			value += 60 * diff;
			value %= (3600 * 24);
			sprintf(buffer, "%02d%02d%02d",
				value / 3600,
				(value % 3600) / 60,
				value % 60);
			return;
		}
		else if(len == 8)
			return;
		now = atol(buffer);
		now += (diff * 60);
		sprintf(buffer, "%10ld", now);
		return;
	}
	else if(!strnicmp(cp, "sec", 3))
	{
		if(len == 6)
		{
			value = atoi(buffer + 4);
			buffer[4] = 0;
			value += atoi(buffer + 2) * 60;
			buffer[2] = 0;
			value += atoi(buffer) * 3600;
			value += diff;
			value %= (3600 * 24);
			sprintf(buffer, "%02d%02d%02d",
				value / 3600,
				(value % 3600) / 60,
				value % 60);
			return;
		}
		else if(len == 8)
			return;
		now = atol(buffer) + diff;
		sprintf(buffer, "%10ld", now);
	}
	if(!strnicmp(cp, "day", 3))
	{
		if(len == 6)
		{
			return;
		}
		else if(len == 8)
		{
			memset(&dtbuff, 0, sizeof(dtbuff));
			dtbuff.tm_mday = atoi(buffer + 6);
			buffer[6] = 0;		
			dtbuff.tm_mon = atoi(buffer + 4) - 1;
			buffer[4] = 0;
			dtbuff.tm_year = atoi(buffer) - 1900;
			now = mktime(&dtbuff);
			now += (60l * 60l * 24l) * (long)diff;
			dt = localtime_r(&now, &dtbuff);
			sprintf(buffer, "%04d%02d%02d",
				dt->tm_year + 1900, dt->tm_mon + 1, dt->tm_mday);
			return;
		}
		now = atol(buffer);
		now += (60l * 60l * 24l) * (long)diff;
		sprintf(buffer, "%10ld", now);
		return;
	}
	else
		sprintf(buffer, "%d", value + diff);
}

ScriptInterp::ScriptInterp(ScriptCommand *cmdset, size_t sym, size_t pg) :
ScriptSymbol(sym, pg)
{
	session = NULL;
	once = true;
	loop = true;
	signalmask = 0;
	packtoken = ',';
	stack = 0;
	cmd = cmdset;
	image = NULL;
	for(tempidx = 0; tempidx < SCRIPT_TEMP_SPACE; ++tempidx)
		temps[tempidx] = (char *)alloc(getSymbolSize() + 1);
	tempidx = 0;
	symsize = sym;
	pgsize = pg;
};

scriptname_t *ScriptInterp::getScriptCopy(const char *name)
{
	char buffer[256];
	scriptname_t *scr;

	snprintf(buffer, 255, "%s::%p", name, this);
	scr = image->dupScript(name, buffer);
	if(scr && !strcmp(script[stack].script->name, name))
		script[stack].script = scr;
	return scr;
}

unsigned long ScriptInterp::getMask(void)
{
	return script[stack].line->mask & script[stack].script->mask;
}

scriptname_t *ScriptInterp::getScriptImage(const char *name)
{
	char buffer[256];
	scriptname_t *scr = image->getScript(name);

	if(!scr)
		return NULL;

	if(scr->mode == scriptname_t::SCRIPT_TYPE_COPIED)
	{
		snprintf(buffer, 255, "%s::%p", name, this);
		scr = image->getScript(buffer);
	}
	return scr;
}

void ScriptInterp::getTrigger(bool jump)
{
	char buffer[256];
	scriptsymbol_t *sym = NULL;
	scriptname_t *scr;
	char *id;

	if(script[stack].local)
		sym = script[stack].local->getTrigger();

	if(sym)
		ScriptSymbol::getTrigger();
	else
		sym = ScriptSymbol::getTrigger();
	if(!jump || !sym)
		return;

	id = strchr(sym->id, '.');
	if(id)
		++id;
	else
		id = sym->id;

	setSymbol("script.trigger", sym->data);
	snprintf(buffer, 255, "%s::%s_%s", script[stack].script->name,
		id, sym->data);

	scr = getScriptImage(buffer);
	if(!scr)
	{
		snprintf(buffer, 256, "%s::%s",  script[stack].script->name, id);
		scr = getScriptImage(buffer);
	}
	if(!scr)
	{
		snprintf(buffer, 256, "%s::%s", id, sym->data);
		scr = getScriptImage(buffer);
	}
	if(scr)
	{
		script[stack].script = scr;
		script[stack].line = scr->first;
		script[stack].index = 0;
	}
}

bool ScriptInterp::scrEnable(void)
{
	char buffer[256];
	char *name = getValue(NULL);
	scriptname_t *scr;
	unsigned long mask, id;
	char *cp;

	if(!name)
	{
		Advance();
		return true;
	}

	if(!strncmp(name, "::", 2))
	{
		strcpy(buffer, script[stack].script->name);
		cp = strstr(buffer, "::");
		if(cp)
			*cp = 0;
		strcat(buffer, name);
		name = buffer;
	}
	scr = getScriptCopy(name);
	if(!scr)
	{
		Error("script-not-found");
		return true;
	}

	while(NULL != (name = getValue(NULL)))
	{
		id = cmd->getTrapId(name);
		mask = cmd->getTrapMask(name);
		if(!mask)
		{
			Error("handler-invalid");
			return true;
		}
		if(!scr->trap[id])
		{
			Error("handler-not-found");
			return true;
		}
		scr->mask |= mask;
	}
	Advance();
	return true;
}

bool ScriptInterp::scrDisable(void)
{
	char buffer[256];
	char *name = getValue(NULL);
	scriptname_t *scr;
	unsigned long mask, id;
	char *cp;

	if(!name)
	{
		Advance();
		return true;
	}

	if(!strncmp(name, "::", 2))
	{
		strcpy(buffer, script[stack].script->name);
		cp = strstr(buffer, "::");
		if(cp)
			*cp = 0;
		strcat(buffer, name);
		name = buffer;
	}
	scr = getScriptCopy(name);
	if(!scr)
	{
		Error("script-not-found");
		return true;
	}

	while(NULL != (name = getValue(NULL)))
	{
		id = cmd->getTrapId(name);
		mask = cmd->getTrapMask(name);
		if(!mask)
		{
			Error("handler-invalid");
			return true;
		}
		if(!scr->trap[id])
		{
			Error("handler-not-found");
			return true;
		}
		scr->mask &= ~mask;
	}
	Advance();
	return true;
}


void ScriptInterp::Advance(void)
{
	script[stack].line = script[stack].line->next;
	if(!script[stack].line && loop)
		script[stack].line = script[stack].script->first;
}

void ScriptInterp::Error(const char *errmsg)
{
	setSymbol("script.error", errmsg);
	if(script[stack].script->mask & 0x02)	
		Trap(1);
	else
		Advance();
}

void ScriptInterp::Trap(const char *trapid)
{
	unsigned trap = cmd->getTrapId(trapid);
	if(!trap)
	{
		if(!stricmp(trapid, "first") || !stricmp(trapid, "top"))
		{
			script[stack].line = script[stack].script->first;
			return;
		}
	}
	Trap(trap);
}

void ScriptInterp::Trap(unsigned id)
{
	scriptline_t *trap = NULL;

	if(getGlobalTrap(id))
		return;

	trap = script[stack].script->trap[id];
	if(!trap && id)
	{
		Advance();
		return;
	}
	script[stack].line = trap;
}	

bool ScriptInterp::Step(const char *trapname)
{
	unsigned long mask, cmask;
	bool rtn;
	scriptsymbol_t *sym = NULL;

	script[stack].index = 0;

	if(session)
	{
		session->waitHandler();
		delete session;
		session = NULL;
	}

	if(trapname)
	{
		getTrigger(false);
		mask = cmd->getTrapMask(trapname);
	}
	else
	{
		mask = 0;
		getTrigger(true);
	}

	// auto return code

retry:
	while(!script[stack].line && stack)
	{
		if(script[stack - 1].local == script[stack].local)
			break;
		Pull();
		if(script[stack].line)
			Advance();
	}
	
	if(!script[stack].line)
	{
		Exit();
		return false;
	}

	cmask = script[stack].line->cmask;
	if(cmask)
	{
		if((cmask & script[stack].script->mask) != cmask)
		{
			Advance();
			goto retry;
		}
	}

	if(mask & script[stack].line->mask)
		Trap(trapname);

	if(script[stack].line->error)
		sym = getEntry("script.error");
	if(sym)
		sym->flags.type = SYM_TYPE_TRIGGER;
	rtn = Execute(script[stack].line->method);
	if(sym)
		sym->flags.type = SYM_TYPE_NORMAL;
	return rtn;
}

bool ScriptInterp::Signal(const char *trapname)
{
	unsigned long mask = cmd->getTrapMask(trapname);
	mask &= script[stack].line->mask;
	mask &= script[stack].script->mask;
	if(!mask)
		return false;

	Stop(mask);
	Trap(trapname);
	return true;
}

bool ScriptInterp::Signal(unsigned id)
{
	if(id >= TRAP_BITS)
		return false;

	unsigned long mask = cmd->getTrapMask(id);
	mask &= script[stack].line->mask;
	if(!mask)
	{
		signalmask |= id;
		return false;
	}

	Stop(mask);
	Trap(id);
	return true;
}

scriptsymbol_t *ScriptInterp::getLocal(const char *id, size_t size)
{
	scriptsymbol_t *sym = NULL;

	if(script[stack].local)
		sym = script[stack].local->getEntry(id, size);

	if(!sym)
		sym = getEntry(id, size);

	return sym;
}		

bool ScriptInterp::scrSkip(void)
{
	int count = atoi(getValue("0"));
	Advance();
	
	while(count--)
		Advance();

	return true;
}

bool ScriptInterp::scrData(void)
{
	Advance();
	return true;
}

bool ScriptInterp::scrSlog(void)
{
	unsigned id = getId();
	const char *member = getMember();
	char *val;

	if(!member)
		member = getKeyword("level");
	
	if(member)
	{
		if(!strnicmp(member, "err", 3))
			slog(SLOG_ERROR);
		else if(!strnicmp(member, "warn", 4))
			slog(SLOG_WARNING);
		else if(!stricmp(member, "debug"))
			slog(SLOG_DEBUG);
		else if(!strnicmp(member, "crit", 4))
			slog(SLOG_CRITICAL);
		else
			slog(SLOG_INFO);
	}
	else
		slog(SLOG_INFO);

	slog() << script[stack].script->name;
	if(id)
		slog() << "(" << id << ")";
	slog() << ": ";
	while(NULL != (val = getValue(NULL)))
		slog() << val;
	slog() << endl;
	Advance();
	return true;
}

bool ScriptInterp::scrUse(void)
{
	Advance();
	return true;
}

bool ScriptInterp::scrMissing(void)
{
	char *chk = getValue(NULL);

	if(!chk)
	{
		Error("missing-no-keyword");
		return true;
	}

	while(chk)
	{
		if(!cmd->getHandler(chk))
			return scrGoto();
		chk = getValue(NULL);
		chk = getValue(NULL);
	}

	Advance();
	return true;
}

bool ScriptInterp::scrHas(void)
{
	char *chk = getValue(NULL);

	if(!chk)
	{
		Error("has-no-keyword");
		return true;
	}

	while(chk)
	{
		if(cmd->getHandler(chk))
			return scrGoto();
		chk = getValue(NULL);
		chk = getValue(NULL);
	}

	Advance();
	return true;
}

bool ScriptInterp::scrSelect(void)
{
	const char *value = getValue(NULL);
	const char *match = getMember();
	int len, l2;
	int offset = 0;

	enum
	{
		SEL_NORMAL,
		SEL_LENGTH,
		SEL_PREFIX,
		SEL_SUFFIX,
		SEL_VALUE
	} sm = SEL_NORMAL;

	if(!value)
	{
		Error("branch-no-select");
		return true;
	}

	if(match)
	{
		if(!strnicmp(match, "len", 3))
			sm = SEL_LENGTH;
		else if(!strnicmp(match, "pre", 3))
			sm = SEL_PREFIX;
		else if(!strnicmp(match, "suf", 3))
			sm = SEL_SUFFIX;
		else if(!strnicmp(match, "val", 3))
			sm = SEL_VALUE;
		else if(!strnicmp(match, "end", 3))
			sm = SEL_SUFFIX;
		else
		{
			offset = atoi(match);
			sm = SEL_PREFIX;
		}
	}
	len = strlen(value);
	if(offset < 0)
		offset = len + offset;
	if(offset < 0)
		offset = 0;
	if(offset > len)
		offset = len;

	while(NULL != (match = getValue(NULL)))
	{
		l2 = strlen(match);
		switch(sm)
		{
		case SEL_SUFFIX:
			if(len < l2)
				break;
			if(!stricmp(match, value + len - l2))
				return scrGoto();
			break;
		case SEL_PREFIX:
			if(!strnicmp(match + offset, value, l2))
				return scrGoto();
			break; 
		case SEL_NORMAL:
			if(!stricmp(match, value))
				return scrGoto();
			break;
		case SEL_LENGTH:
			if(atoi(match) == len)
				return scrGoto();
			break;
		case SEL_VALUE:
			if(atoi(match) == atoi(value))
				return scrGoto();
			break;
		}
		getValue(NULL);
	}
	Advance();
	return true;
}		

bool ScriptInterp::scrOnce(void)
{
	if(getOnce())
		return scrGoto();
	Advance();
	return true;
}

bool ScriptInterp::scrOn(void)
{
	char *sig = getValue(NULL);
	unsigned long mask;

	if(!sig)
	{
		Error("on-no-signal");
		return true;
	}

	mask = cmd->getTrapMask(sig);
	if(!mask)
	{
		Error("on-invalid-signal");
		return true;
	}
	if(signalmask & mask)
	{
		signalmask &= ~mask;
		return scrGoto();
	}
	Advance();
	return true;
}

bool ScriptInterp::scrGoto(void)
{
	char namebuf[256];
	char *label = getValue(NULL);
	char *ext;
	scriptline_t *skip;
	scriptname_t *scr;

	if(!label)
	{
		Error("branch-failed");
		return true;
	}

	if(*label == '^')
	{
		initKeywords(0);
		Trap(++label);
		return true;
	}

	if(*label >= '0' && *label <= '9' && !label[1])
	{
		initKeywords(0);
		scriptline_t *skip = script[stack].script->skip[atoi(label)];
		if(!skip)
			Advance();

		return true;
	}

	if(!strncmp(label, "::", 2))
	{
		strcpy(namebuf, script[stack].script->name);
		ext = strstr(namebuf, "::");
		if(ext)
			strcpy(ext, label);		
		else
			strcat(namebuf, label);
		label = namebuf;
	}
	scr = getScriptImage(label);
	if(!scr)
	{
		Error("script-not-found");
		return true;
	}
	initKeywords(0);
	once = true;
	script[stack].script = scr;
	script[stack].line = scr->first;
	script[stack].index = 0;
	return true;
}

bool ScriptInterp::scrTry(void)
{
	char namebuf[256];
	scriptname_t *scr;	
	char *label, *ext;

	while(NULL != (label = getValue(NULL)))
	{
		if(!strncmp(label, "::", 2))
		{
			strcpy(namebuf, script[stack].script->name);
			ext = strstr(namebuf, "::");
			if(ext)
				strcpy(ext, label);
			else		
				strcat(namebuf, label);			
			label = namebuf;
		}
		scr = getScriptImage(label);
		if(!scr)
			continue;
		once = true;
		script[stack].script = scr;
		script[stack].line = scr->first;
		script[stack].index = 0;
		return true;
	}
	Advance();
	return true;
}

bool ScriptInterp::scrArm(void)
{
	scriptsymbol_t *opt;

	while(opt = getVariable(0))
	{
		if(opt->flags.type == SYM_TYPE_NORMAL)
			opt->flags.type = SYM_TYPE_TRIGGER;
	}
	Advance();
	return true;
}

bool ScriptInterp::scrDisarm(void)
{
	scriptsymbol_t *opt;

	while(opt = getVariable(0))
	{
		if(opt->flags.type == SYM_TYPE_TRIGGER)
			opt->flags.type = SYM_TYPE_NORMAL;
	}
	Advance();
	return true;
}

bool ScriptInterp::scrSwap(void)
{
	scriptsymbol_t *s1, *s2;
	s1 = getVariable(getSymbolSize());
	s2 = getVariable(getSymbolSize());
	char *id;

	if(!s1 || !s2)
	{
		Error("symbol-not-found");
		return true;
	}
	EnterMutex();
	id = s1->id;
	s1->id = s2->id;
	s2->id = id;
	LeaveMutex();
	Advance();
	return true;
}

bool ScriptInterp::scrRead(void)
{
	scriptsymbol_t *sym;
	scriptname_t *scr;
	scriptline_t *rd;
	const char *mem = getMember();
	int argc = 0;
	char *value = NULL;
	char namebuf[256];

	if(!mem)
		mem = "";
	if(!stricmp(mem, "from"))
	{
		value = getValue(script[stack].script->name);
		if(!strnicmp(value, "::", 2))
		{
			strcpy(namebuf, script[stack].script->name);
			strcat(namebuf, value);
			value = namebuf;
		}
		scr = getScriptImage(value);
		if(!scr)
		{
			Error("no-source-to-read");
			return true;
		}
		script[stack].read = scr->first;
		value = getValue(NULL);
		Advance();
		return true;
	}
	rd = script[stack].read;
	while(rd)
	{
		if(rd->method == &ScriptInterp::scrData)
			break;
		rd = rd->next;
	}
	if(!rd)
	{
		script[stack].read = NULL;
		Error("end-of-data");
		return true;
	} 
	else
		script[stack].read = rd->next;
	while(argc < rd->argc)
	{
		value = getContent(rd->args[argc++]);
		if(!value)
			break;
		sym = getVariable(strlen(value));
		if(!sym)
			break;
		if(sym->flags.readonly)
			continue; 
		strcpy(sym->data, value);
		sym->flags.initial = false;
		if(sym->flags.commit)
			Commit(sym);
	}
	
	Advance();
	return true;
}				
	
bool ScriptInterp::scrRepeat(void)
{
	unsigned short loop = script[stack].line->loop;
	scriptline_t *line;
	int index = script[stack].index;
	int count;

	script[stack].index = 0;
	count = atoi(getValue("0"));

	if(index >= count)
	{
		line = script[stack].line->next;
		while(line)
		{
			if(line->loop == loop)
			{
				script[stack].line = line;
				Advance();
				return true;
			}
			line = line->next;
		}
		Error("loop-overflow");
		return true;		
	}
	script[stack].index = ++index;
	if(!Push())
		return true;

	Advance();
	return true;	
}

bool ScriptInterp::scrForeach(void)
{
	scriptsymbol_t *sym;
	unsigned short loop = script[stack].line->loop;
	scriptline_t *line;
	int index = script[stack].index;
	const char *value = getMember();
	int len = 0;

	script[stack].index = 0;

	sym = getVariable(getSymbolSize());

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	sym->data[sym->flags.size] = 0;

	if(value && !index)
		len = atoi(value);		

	value = getValue();
	while(value[index] && (len--) > 1)
	{
		while(value[index] && value[index] != packtoken)
			++index;
		if(value[index] == packtoken)
			++index;
	}
	
	if(!value)
	{
		Error("list-missing");
		return true;
	}

	if(!value[index])
	{
		line = script[stack].line->next;
		while(line)
		{
			if(line->loop == loop)
			{
				script[stack].line = line;
				Advance();
				return true;
			}
			line = line->next;
		}
		Error("loop-overflow");
		return true;
	}
	len = 0;
	while(value[index] && value[index] != packtoken && len < sym->flags.size)
		sym->data[len++] = value[index++];
	if(value[index] == packtoken)
		++index;
	sym->data[len] = 0;
	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);

	script[stack].index = index;
	if(!Push())
		return true;

	Advance();
	return true;
}

bool ScriptInterp::scrFor(void)
{
	scriptsymbol_t *sym;
	unsigned short loop = script[stack].line->loop;
	scriptline_t *line;
	int index = script[stack].index;
	script[stack].index = 0;
	
	if(!index)
		++index;

	sym = getVariable(getSymbolSize());
	char *value;
	
	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	sym->data[sym->flags.size] = 0;
	script[stack].index = index;
	value = getValue(NULL);
	if(!value)
	{
		line = script[stack].line->next;
		while(line)
		{
			if(line->loop == loop)
			{
				script[stack].line = line;
				Advance();
				return true;
			}
			line = line->next;
		}
		Error("loop-overflow");
		return true;				
	}

	strncpy(sym->data, value, sym->flags.size);
	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);

	if(!Push())
		return true;

	Advance();
	return true;
}	
	
bool ScriptInterp::scrDo(void)
{
	unsigned short loop = script[stack].line->loop;
	scriptline_t *line;
		
	script[stack].index = 0;	// always reset

	if(script[stack].line->argc)
	{
		if(!Conditional())
		{
			line = script[stack].line->next;
			while(line)
			{
				if(line->loop == loop)
				{
					script[stack].line = line;
					Advance();
					return true;
				}
				line = line->next;
			}
			Error("loop-overflow");
			return true;
		}
	}

	if(!Push())
		return true;

	Advance();
	return true;
}
	
bool ScriptInterp::scrLoop(void)
{
	unsigned short loop;

	if(stack < 1)
	{
		Error("stack-underflow");
		return true;
	}

	loop = script[stack - 1].line->loop;
	if(!loop)
	{
		Error("stack-not-loop");
		return true;
	}
	
	if(script[stack].line->argc)
	{
		if(!Conditional())
		{
			script[stack - 1] = script[stack];
			--stack;
			Advance();
			return true;
		}
	}
	
	--stack;
	return Execute(script[stack].line->method);
}

bool ScriptInterp::scrContinue(void)
{
	scriptline_t *line;
	unsigned short loop;

	if(script[stack].line->argc)
	{
		if(!Conditional())
		{
			Advance();
			return true;
		}
	}

	if(stack < 1)
	{
		Error("stack-underflow");
		return true;
	}

	loop = script[stack - 1].line->loop;
	line = script[stack].line->next;
	
	if(!loop)
	{
		Error("stack-not-loop");
		return true;
	}

	while(line)
	{
		if(line->loop == loop)
		{
			script[stack].line = line;	
			return true;
		}
		line = line->next;
	}
	Error("loop-overflow");
	return true;
}

bool ScriptInterp::scrBreak(void)
{
	scriptline_t *line;
	unsigned short loop;

	if(script[stack].line->argc)
	{
		if(!Conditional())
		{
			Advance();
			return true;
		}
	}

	if(stack < 1)
	{
		Error("stack-underflow");
		return true;
	}
	
	loop = script[stack - 1].line->loop;
	line = script[stack].line->next;

	if(!loop)
	{
		Error("stack-not-loop");
		return true;
	}

	while(line)
	{
		if(line->loop == loop)
		{
			--stack;
			script[stack].line = line;
			Advance();
			return true;
		}
		line = line->next;
	}
	Error("loop-overflow");
	return true;
}

bool ScriptInterp::scrIfThen(void)
{
	if(!Conditional())
		Advance();
	Advance();
	return true;
}

bool ScriptInterp::scrIf(void)
{
	if(Conditional())
		return scrGoto();

	Advance();
	return true;
}

bool ScriptInterp::scrLock(void)
{
	unsigned id = atoi(getValue("0"));
	const char *member = getMember();

	if(!member)
		member = "wait";

	if(!stricmp(member, "try"))
	{
		if(!locks.Lock(this, id))
			Error("lock-busy");
		else
			Advance();
		return true;
	}

	if(!stricmp(member, "unlock") || !stricmp(member, "ulock"))
	{
		if(!locks.Unlock(this, id))
			Error("lock-invalid");
		else
			Advance();
		return true;
	}

	if(!locks.Lock(this, id))
		return true;

	Advance();
	return true;
}

bool ScriptInterp::scrCall(void)
{
	int index = script[stack].index;
	int id = 0;
	scriptsymbol_t *sym;
	char symname[8];
	char *arg;

	if(!Push())
		return true;

	script[stack].local = new ScriptSymbol(symsize, pgsize);

	while(NULL != (arg = getValue(NULL)))
	{
		sprintf(symname, "%d", id++);
		sym = getLocal(symname, pgsize);
		if(sym)
		{
			sym->flags.initial = false;
			strcpy(sym->data, arg);
		} 
	}
	
	script[stack].index = index;
	return scrGoto(); 
}

bool ScriptInterp::scrPop(void)
{
	char *level = getValue(NULL);
	int max;

	if(level)
	{
		max = atoi(level);
		if(max < stack)
			stack = max;
		Advance();
		return true;
	}

	if(stack < 1)
	{
		Error("stack-underflow");
		return true;
	}
	if(script[stack].local != script[stack - 1].local)
	{
		if(script[stack].local)
			delete script[stack].local;
		script[stack].local = script[stack - 1].local;
	}
	script[stack - 1] = script[stack];
	--stack;
	Advance();
	return true;
}

bool ScriptInterp::scrReturn(void)
{
	bool error;

	error = false;
	do {
		if (stack == 0) {
			Error("stack-underflow");
			error = true;
		} else {
			stack--;
		}
	} while(!error && script[stack].line->loop != 0);

	if (!error) {
	Advance();
	}

	return true;
}
		
bool ScriptInterp::scrExit(void)
{
	while(stack)
		Pull();

	script[stack].line = NULL;
	return true;
}

bool ScriptInterp::scrClear(void)
{
	scriptsymbol_t *sym = getVariable(0);

	while(sym)
	{
		if(sym->flags.type == SYM_TYPE_FIFO || sym->flags.type == SYM_TYPE_SEQUENCE || sym->flags.type == SYM_TYPE_STACK)
		{
			sym->data[1] = sym->data[2] = 0;
			continue;
		}
		if(sym->flags.readonly && !sym->flags.commit)
		{
			sym = getVariable(0);
			continue;
		}
		sym->data[0] = 0;
		sym->flags.initial = true;
		if(sym->flags.commit)
			Commit(sym);
		sym = getVariable(0);
	}
	Advance();
	return true;
}

bool ScriptInterp::scrPost(void)
{
	scriptsymbol_t *sym;
	char *opt = getOption(NULL);
	if(!opt)
	{
		Error("symbol-missing");
		return true;
	}

	if(*opt != '%')
	{
		Error("symbol-invalid");
		return true;
	}

	sym = getLocal(++opt, 0);
	if(!sym)
	{
		Error("symbol-missing");
		return true;
	}

	if(sym->flags.type != SYM_TYPE_FIFO && sym->flags.type != SYM_TYPE_SEQUENCE && sym->flags.type != SYM_TYPE_STACK)
	{
		Error("symbol-type-invalid");
		return true;
	}

	while(NULL != (opt = getValue(NULL)))
		postSymbol(sym, opt);

	Advance();
	return true;
}
	
bool ScriptInterp::scrList(void)
{
	unsigned count = 0;
	scriptsymbol_t *sym;
	const char *opt;
	const char *name = getOption(NULL);
	char def[96], buf[8];

	if(!name)
	{
		Error("list-missing");
		return true;
	}

	if(*name != '%')
	{
		Error("list-missing-array-name");
		return true;
	}
	++name;

	strcpy(def, name);
	strcat(def, ".#####");
	if(!setAlias(name, def))
	{
		Error("array-unavailable");
		return true;
	}

	while(NULL != (opt = getOption(NULL)))
	{
		++count;
		snprintf(def, sizeof(def), "%s.%d", name, count);
		setConst(def, opt);
	}
	sprintf(buf, "%d", count);
	strcpy(def, name);
	strcat(def, ".count");
	setConst(def, buf);
	strcpy(def, name);
	strcat(def, ".limit");
	setConst(def, buf);
	strcpy(def, name);
	strcat(def, ".index");
	sym = getEntry(def, 5);
	if(!sym)
	{
		Error("array-no-index");
		return true;
	}
	sym->flags.initial = false;
	sym->flags.commit = true;
	sym->flags.type = SYM_TYPE_INDEX;
	strcpy(sym->data, "0");
	Advance();
	return true;
}

bool ScriptInterp::scrArray(void)
{
	unsigned size = getSymbolSize();
	unsigned count = atoi(getValue("0"));
	const char *mem = getMember();
	char def[96], buf[8];
	unsigned index;
	scriptsymbol_t *sym;

	if(mem)
		size = atoi(mem);

	if(!count || !size)
	{
		Error("array-no-size");
		return true;
	}

	while(NULL != (mem = getOption(NULL)))
	{
		strcpy(def, mem);
		strcat(def, ".#####");
		if(!setAlias(mem, def))
		{
			Error("array-unavailable");
			return true;
		}

		snprintf(def, sizeof(def), "%s.index", mem);
		sym = getEntry(def, 5);
		if(!sym)
		{
			Error("array-no-index");
			return true;
		}
		sym->flags.initial = false;
		sym->flags.commit = true;
		sym->flags.type = SYM_TYPE_INDEX;
		strcpy(sym->data, "0");
					
		sprintf(buf, "%d", count);
		snprintf(def, sizeof(def), "%s.count", mem);
		setConst(def, buf);
		snprintf(def, sizeof(def), "%s.limit", mem);
		setConst(def, buf);
		
		for(index = 1; index <= count; ++index)
		{
			snprintf(def, sizeof(def), "%s.%d", mem, index);
			setSymbol(def, size);	
		}	
	 }
	Advance();
	return true;
}

bool ScriptInterp::scrFifo(void)
{
	unsigned char rec = getSymbolSize();
	unsigned char count = atoi(getValue("0"));
	const char *mem = getMember();
	int rtn;
	if(mem)
		rec = atoi(mem);

	if(!count || !rec)
	{
		Error("symbol-no-size");
		return true;
	}

	while(NULL != (mem = getOption(NULL)))
	{
		if(strchr(mem, '.') || !script[stack].local)
			rtn = makeFifo(mem, count, rec);
		else
			rtn = script[stack].local->makeFifo(mem, count,  rec);
		if(!rtn)
		{
			Error("fifo-make-failed");
			return true;
		}
	}
	Advance();
	return true;
}

bool ScriptInterp::scrCounter(void)
{
	const char *id;
	int rtn;

	while(NULL != (id = getOption(NULL)))
	{
		if(strchr(id, '.') || !script[stack].local)
			rtn = makeCounter(id);
		else
			rtn = script[stack].local->makeCounter(id);
		if(!rtn)
		{
			Error("counter-make-failed");
			return true;
		}
	}
	Advance();
	return true;
}

bool ScriptInterp::scrStack(void)
{
	unsigned char rec = getSymbolSize();
	unsigned char count = atoi(getValue("0"));
	const char *mem = getMember();
	if(mem)
		rec = atoi(mem);

	if(!count || !rec)
	{
		Error("symbol-no-size");
		return true;
	}

	while(NULL != (mem = getOption(NULL)))
	{
		if(!makeStack(mem, count, rec))
		{
			Error("stack-make-failed");
			return true;
		}
	}
	Advance();
	return true;
}


bool ScriptInterp::scrSequence(void)
{
	unsigned char rec = getSymbolSize();
	unsigned char count = atoi(getValue("0"));
	const char *mem = getMember();
	int rtn;
	if(mem)
		rec = atoi(mem);

	if(!count || !rec)
	{
		Error("symbol-no-size");
		return true;
	}
	
	mem = getOption(NULL);
	if(!mem)
	{
		Error("symbol-missing");
		return true;
	}

	if(strchr(mem, '.') || !script[stack].local)
		rtn = makeSequence(mem, count, rec);
	else
		rtn = script[stack].local->makeSequence(mem, count, rec);

	if(!rtn)
	{
		Error("sequence-make-failed");
		return true;
	}
	--script[stack].index;
	scrPost();
}


bool ScriptInterp::scrSize(void)
{
	scriptsymbol_t *sym;
	int size = atoi(getValue("0"));
	char *opt = getOption(NULL);

	if(!opt)
	{
		Error("symbol-not-specified");
		return true;
	}

	if(!size)
	{
		Error("symbol-no-size");
		return true;
	}

	while(opt)
	{
		if(*opt == '@')
		{
			opt = getSymbol(++opt);
			if(!opt)
			{
				opt = getOption(NULL);
				continue;
			}
		}
		else if(*opt != '%')
		{
			Error("symbol-not-constant");
			return true;
		}
	
		if(*opt == '%')
			++opt;
		if(strchr(opt, '.') || !script[stack].local)
			sym = getEntry(opt, size);
		else
			sym = script[stack].local->getEntry(opt, size);
		opt = getOption(NULL);
	}

	Advance();
	return true;
}

bool ScriptInterp::scrInc(void)
{
	scriptsymbol_t *sym = getVariable(getSymbolSize());
	int len;
	int value = 0;
	char *opt;
	struct tm *dt, dtbuff;
	time_t now;

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	len = sym->flags.size;
	opt = getValue(NULL);
	if(opt)
	{
		while(opt)
		{
			value += atoi(opt);
			opt = getValue(NULL);
		}
	}
	else
		++value;


	adjustValue(sym->data, getMember(), value);
	if(sym->flags.commit)
		Commit(sym);

	Advance();
	return true;
}

bool ScriptInterp::scrDec(void)
{
	scriptsymbol_t *sym = getVariable(getSymbolSize());
	int len;
	int value = 0;
	char *opt;

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	len = sym->flags.size;
	opt = getValue(NULL);
	if(opt)
	{
		while(opt)
		{
			value -= atoi(opt);
			opt = getValue(NULL);
		}
	}
	else
		--value;

	adjustValue(sym->data, getMember(), value); 
	if(sym->flags.commit)
		Commit(sym);

	Advance();
	return true;
}

bool ScriptInterp::scrTryeach(void)
{
	char *gp = getValue(NULL);
	const char *mem = getMember();
	int offset = 0;
	char namebuf[256];
	char *np;
	scriptname_t *scr;

	if(mem)
		offset = atoi(mem);

	if(!gp)
	{
		Error("list-not-found");		
		return true;
	}

	while(*gp && (offset--) > 1)
	{
		gp = strchr(gp, packtoken);
		if(gp)
			++gp;
		else
			gp = "";
	}

	while(*gp)
	{
		np = namebuf;
		while(*gp && *gp != packtoken)
			*(np++) = *(gp++);
	
		*np = 0;
		if(*gp == packtoken)
			++gp;

		scr = getScriptImage(namebuf);
		if(!scr)
			continue;
	
		once = true;
		script[stack].script = scr;
		script[stack].line = scr->first;		
		script[stack].index = 0;
		return true;
	}
	Advance();
	return true;
}
	
bool ScriptInterp::scrUnpack(void)
{
	scriptsymbol_t *sym;
	char *pp;
	char *name;
	char *gp = getValue(NULL);
	int len;
	const char *mem = getMember();
	int offset = 0;

	if(mem)
		offset = atoi(mem);

	if(!gp)
	{
		Error("source-not-found");
		return true;
	}

	while(*gp && (offset--) > 1)
	{
		gp = strchr(gp, packtoken);
		if(!gp)
			gp = "";
		else
			++gp;
	}

	while(*gp && NULL != (name = getOption(NULL)))
	{
		if(*name == '@')
			sym = getIndirect(++name);
		if(*name == '%')
			sym = getEntry(++name, getSymbolSize());
		else
			sym = NULL;

		if(sym)
		{
			if(sym->flags.readonly)
				pp = NULL;
			else
				pp = sym->data;
		}
		else
			pp = NULL;

		len = 0;
		while(pp && *gp && *gp != packtoken)
		{
			if(++len > sym->flags.size)
				break;

			*(pp++) = *(gp++);
		}

		if(pp)
		{
			*pp = 0;
			sym->flags.initial = false;
			if(sym->flags.commit)
				Commit(sym);
		}

		while(*gp && *gp != packtoken)
			++gp;

		if(*gp == packtoken)
			++gp;
	}
	Advance();
	return true;
}		

bool ScriptInterp::scrPack(void)
{
	scriptsymbol_t *sym;
	char *name = getOption(NULL);
	char *value;
	int len = 0;
	char *pp;

	if(!name)
	{
		Error("symbol-not-found");
		return true;
	}

	if(*name == '@')
		sym = getIndirect(++name);
	else if(*name == '%')
		sym = getEntry(++name, getSymbolSize());

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-read-only");
		return true;
	}

	pp = sym->data;
	len = strlen(pp);
	pp += len;
		
	value = getValue(NULL);
	while(len < sym->flags.size && value != NULL)
	{
		if(len)
		{
			*(pp++) = packtoken;
			++len;
		}
		while(*value && len++ < sym->flags.size)
			*(pp++) = *(value++);

		value = getValue(NULL);
	}
	*pp = 0;
	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);
	Advance();
	return true;
}

bool ScriptInterp::scrPacked(void)
{
	char *value = getValue(NULL);
	if(value)
		packtoken = *value;
	Advance();
	return true;
}

bool ScriptInterp::scrLoadable(void)
{
	scriptline_t *line = getScript();
	char *cmd = line->cmd;
	char keybuf[33];
	int len = 0;
	char *kw = keybuf;
	ScriptModule *mod = ScriptModule::first;

	while(len ++ < 32 && *cmd && *cmd != '.')
		*(kw++) = *(cmd++);
	*kw = 0;

	while(mod)
	{
		if(!stricmp(keybuf, mod->cmd))
			break;
		mod = mod->next;
	}
	
	if(!mod)
	{
		Error("module-not-found");
		return true;
	}
	kw = mod->getSession(this, line, &session);
	if(kw)
	{
		Error(kw);
		return true;
	}
	if(!session)
	{
		Advance();
		return true;
	}
	return false;
}

bool ScriptInterp::scrToday(void)
{
	scriptsymbol_t *sym = getVariable(11);
	const char *format = getMember();
	char *origin = getValue(NULL);
	struct tm dtbuf, *dt;
	time_t now;
	int len;
	char save[10];

	if(!sym)
	{
		Error("no-symbol");
		return true;
	}

	time(&now);
	if(origin)
	{
		len = strlen(origin);
		if(len == 6)
		{
			strcpy(save, origin);
			now /= 3600 * 24;
			now *= 3600 * 24;
			now += atoi(save + 4);
			save[4] = 0;
			now += atoi(save + 2) * 60;
			save[2] = 0;
			now += atoi(save) * 3600;
		}
		else if(len == 8)
		{
			strcpy(save, origin);
			memset(&dtbuf, 0, sizeof(dtbuf));
			dtbuf.tm_mday = atoi(save + 6);
			save[6] = 0;
			dtbuf.tm_mon = atoi(save + 4) - 1;
			save[4] = 0;
			dtbuf.tm_year = atoi(save) - 1900;
			now = mktime(&dtbuf);
		}
		else
			now = atol(origin);
	}

	dt = localtime_r(&now, &dtbuf);
	if(!format)
		format = "unix";

	if(!stricmp(format, "date"))
		sprintf(sym->data, 
			"%04d%02d%02d", dt->tm_year + 1900, dt->tm_mon + 1, dt->tm_mday);
	else if(!stricmp(format, "time"))
		sprintf(sym->data, "%02d%02d%02d", dt->tm_hour, dt->tm_min, dt->tm_sec);
	else if(!stricmp(format, "hour"))
		sprintf(sym->data, "%d", dt->tm_hour);
	else
		sprintf(sym->data, "%010ld", now);

	if(sym->flags.commit)
		Commit(sym);

	Advance();
	return true;
}

bool ScriptInterp::scrAlias(void)
{
	char *sym;
	char *src;

	while(NULL != (sym = getValue(NULL)))
	{
		src = getValue(NULL);
		if(!src)
		{
			Error("alias-no-source");
			return true;
		}

		if(!setAlias(sym, src))
		{
			Error("alias-failure");
			return true;
		}
	}

	Advance();
	return true;
}

bool ScriptInterp::scrConst(void)
{
	char *sym = getValue(NULL);
	char *val;
	char buffer[getSymbolSize()];
	strstream str(buffer, getSymbolSize());
	int count = 0;
	int idx = 0;
	scriptline_t *line = getScript();

	buffer[0] = 0;
	if(!sym)
	{
		while(idx < line->argc)
		{
			val = line->args[idx++];
			if(*val != '=')
				continue;
			if(*(++val) == '%')
				++val;
			setConst(val, line->args[idx++]);
			++count;
		}
		
		if(count)
			Advance();
		else
			Error("const-not-specified");
		return true;
	}

	while(NULL != (val = getValue(NULL)))
		str << val;
	str << ends;
	if(!setConst(sym, buffer))
	{
		Error("const-not-set");
		return true;
	}
	Advance();
	return true;
}
	
bool ScriptInterp::scrDup(void)
{
	char *id;
	scriptsymbol_t *src, *dup;
	while(NULL != (src = getVariable(0)))
	{
		dup = getVariable(src->flags.size);
		if(!dup)
		{
			Error("no-target");
			return true;

		}
		EnterMutex();
		if(!dup->flags.initial)
		{
			LeaveMutex();
			Error("target-exists");
			return true;
		}
		id = dup->id;
		memcpy(dup, src, sizeof(scriptsymbol_t) + src->flags.size);
		dup->id = id;
		LeaveMutex();
	}
	Advance();
	return true;
}
	
bool ScriptInterp::scrInit(void)
{
	ScriptProperty *prop = NULL;
	scriptsymbol_t *sym;
	int len;
	const char *value = getMember();
	int size = 0;
	bool number = false;
	int count = 0;
	int idx = 0;
	scriptline_t *line = getScript();

	enum
	{
		VAL_NONE,
		VAL_MIN,
		VAL_MAX
	}	minmax = VAL_NONE;

	if(value)
	{
		if(!stricmp(value, "size"))
			size = atoi(getValue("0"));
		else if(!stricmp(value, "max"))
			minmax = VAL_MAX;
		else if(!stricmp(value, "min"))
			minmax = VAL_MIN;
		else if(!stricmp(value, "val") || !stricmp(value, "num"))
		{
			size = 11;
			number = true;
		}
		else if(NULL == (prop = ScriptProperty::find(value)))
			size = atoi(value);
	}

	if(!size)
		size = getSymbolSize();

	sym = getVariable(size);
	if(!sym)
	{
		while(idx < line->argc)
		{
			value = line->args[idx++];
			if(*value != '=')
				continue;
			if(*(++value) == '%')
				++value;
			++count;
			sym = getEntry(value, size);
			value = line->args[idx++];
			if(!sym)
				continue;
			if(!sym->flags.initial || sym->flags.readonly)
				continue;
			setSymbol(sym->id, value);
		}
		if(count)
			Advance();
		else
			Error("symbol-not-found");
		return true;
	}

	if(!sym->flags.initial)
	{
		Advance();
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	sym->data[sym->flags.size] = 0;
	len = 0;

	while(len < sym->flags.size && NULL != (value = getValue(NULL)))
	{
		if(minmax == VAL_MIN && atoi(value) >= atoi(sym->data) && len)
			continue;
		if(minmax == VAL_MAX && atoi(value) <= atoi(sym->data) && len)
			continue;
		if(minmax == VAL_NONE)
			strncpy(sym->data + len, value, sym->flags.size - len);
		else
			strncpy(sym->data, value, sym->flags.size);
		sym->data[sym->flags.size] = 0;
		len = strlen(sym->data);	
	}

	if(number)
		sprintf(sym->data, "%ld", atol(sym->data));
	else if(prop)
	{
		strcpy(temps[tempidx], sym->data);
		prop->setProperty(sym->data, temps[tempidx],  sym->flags.size);
	}
	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);

	Advance();
	return true;
}

bool ScriptInterp::scrSet(void)
{
	ScriptProperty *prop = NULL;
	scriptsymbol_t *sym;
	int size = 0;
	int len;
	char *cp, *ep;
	const char *value;
	bool number = false;

	enum
	{
		VAL_NONE,
		VAL_MIN,
		VAL_MAX
	}	minmax = VAL_NONE;

	value = getMember();
	if(value)
	{
		if(!stricmp(value, "size"))
			size = atoi(getValue("0"));
		else if(!stricmp(value, "min"))
			minmax = VAL_MIN;
		else if(!stricmp(value, "max"))
			minmax = VAL_MAX;
		else if(!strnicmp(value, "val", 3) || !strnicmp(value, "num", 3))
		{
			number = true;
			size = 11;
		}
		else if(NULL ==(prop = ScriptProperty::find(value)))
			size = atoi(value);
	}
	if(!size)
		size = getSymbolSize();
	sym = getVariable(size);
	if(!sym)
	{
		if(!initKeywords(size))
			Error("symbol-not-found");
		else
			Advance();			
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	sym->data[sym->flags.size] = 0;
	len = 0;

	while(len < sym->flags.size && NULL != (value = getValue(NULL)))
	{
		if(minmax == VAL_MIN && atoi(value) >= atoi(sym->data) && len)
			continue;

		if(minmax == VAL_MAX && atoi(value) <= atoi(sym->data) && len)
			continue;
		if(minmax == VAL_NONE)
			strncpy(sym->data + len, value, sym->flags.size - len);
		else
			strncpy(sym->data, value, sym->flags.size);
		sym->data[sym->flags.size] = 0;
		len = strlen(sym->data);	
	}

	if(number)
		sprintf(sym->data, "%ld", atol(sym->data));
	else if(prop)
	{
		strcpy(temps[tempidx], sym->data);
		prop->setProperty(sym->data, temps[tempidx], sym->flags.size);
	}

	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);

	Advance();
	return true;
}


bool ScriptInterp::scrGather(void)
{
	scriptsymbol_t *sym = getVariable(getSymbolSize());
	char *suffix = getValue(NULL);
	scriptname_t *list[33];
	int count = 0, idx;
	char *ext;
	scriptname_t *scr;
	
	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}
	
	count = image->Gather(suffix, list, 32);
	if(!count)
	{
		Error("no-scripts-found");
		return true;
	}
	sym->data[0] = 0;
	for(idx = 0; idx < count; ++idx)
	{
		if(idx)
			strcat(sym->data, ",");
		strcat(sym->data, list[idx]->name);
	}
	sym->flags.initial = false;
	if(sym->flags.commit)	
		Commit(sym);
	Advance();
	return true;
}			

bool ScriptInterp::scrFullpath(void)
{
	char *cp;
	char *pro;
	int len;
	scriptsymbol_t *sym = getVariable(getSymbolSize());
	char *prefix = getValue(NULL);

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	if(!prefix)
	{
		Error("no-prefix");
		return true;
	}

	cp = strchr(sym->data, '/');
	pro = strchr(sym->data, ':');
	if(cp == sym->data || pro < cp)
	{
		Advance();
		return true;
	}
	len = strlen(prefix);
	if(sym->flags.size < len + strlen(sym->data) + 3)
	{
		Error("no-space");
		return true;
	}
	cp = sym->data;
	sym->data[len++] = '/';
	while(*cp)
		sym->data[len++] = *(cp++);
	*cp = 0;
	memcpy(sym->data, prefix, strlen(prefix));
	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);
	Advance();
	return true;
}
		
bool ScriptInterp::scrDirname(void)
{
	char *cp;
	scriptsymbol_t *sym = getVariable(1);

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	cp = strrchr(sym->data, '/');
	if(!cp)
		strcpy(sym->data, ".");
	else
		*cp = 0;

	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);

	Advance();
	return true;
}

bool ScriptInterp::scrBasename(void)
{
	const char *ext;
	scriptsymbol_t *sym = getVariable(0);
	char *cp;

	if(!sym)
	{
		Error("symbol-not-found");
		return true;
	}

	if(sym->flags.readonly)
	{
		Error("symbol-readonly");
		return true;
	}

	cp = strrchr(sym->data, '/');
	if(cp)
		strcpy(sym->data, cp + 1);
	
	while(NULL != (ext = getValue(NULL)))
	{
		cp = sym->data + strlen(sym->data) - strlen(ext) - 1;
		if(cp > sym->data)
			if(*cp == '.')
				if(!stricmp(cp + 1, ext))
					*cp = 0;
	}
	sym->flags.initial = false;
	if(sym->flags.commit)
		Commit(sym);
	Advance();
	return true;
}

bool ScriptInterp::Conditional(void)
{
	char *v1 = getOption(NULL);
	char *op = getValue(NULL);
	char *v2 = getOption(NULL);
	int l1, l2;

	// no first parm, invalid

	if(!v1)
		return false;
	
	v1 = getContent(v1);

	// sym/label by itself, invalid

	if(!op)
	{
		script[stack].index = 0;
		return false;
	}
	
	// ifdef sym ... format assumed

	if(!v2)
	{
		script[stack].index = 1;
		if(v1 && *v1)
			return true;
		return false;
	}
	
	v2 = getContent(v2);

	if(!v1)
		v1 = "";

	if(!v2)
		v2 = "";
	
	if(!stricmp(op, "=") || !stricmp(op, "-eq"))
	{
		if(atol(v1) == atol(v2))
			return true;
		return false;
	}

	if(!stricmp(op, "<>") || !stricmp(op, "-ne"))
	{
		if(atol(v1) != atol(v2))
			return true;
		return false;
	}

	if(!stricmp(op, "==") || !stricmp(op, ".eq."))
	{
		if(!stricmp(v1, v2))
			return true;

		return false;
	}

	if(!stricmp(op, "!=") || !stricmp(op, ".ne."))
	{
		if(stricmp(v1, v2))
			return true;

		return false;
	}

	if(!stricmp(op, "$"))
	{
		if(strstr(v2, v1))
			return true;
		return false;
	}

	if(!stricmp(op, "$<") || !stricmp(op, "$+"))
	{
		if(!strnicmp(v1, v2, strlen(v1)))
			return true;
		return false;
	}

	if(!stricmp(op, "$>") || !stricmp(op, "$-"))
	{
		l1 = strlen(v1);
		l2 = strlen(v2);
		if(l1 <= l2)
			if(!strnicmp(v1, v2 + l2 - l1, l1))
				return true;

		return false;
	}

	if(!stricmp(op, "<") || !stricmp(op, "-lt"))
	{
		if(atol(v1) < atol(v2))
			return true;
		return false;
	}

	if(!stricmp(op, "<=") || !stricmp(op, "=<") || !stricmp(op, "-le"))
	{
		if(atol(v1) <= atol(v2))
			return true;
		return false;
	}

	if(!stricmp(op, ">") || !stricmp(op, "-gt"))
	{
		if(atol(v1) > atol(v2))
			return true;
		return false;
	}

	if(!stricmp(op, ">=") || !stricmp(op, "=>") || !stricmp(op, "-ge"))
	{
		if(atol(v1) >= atol(v2))
			return true;
		return false;
	}

	// if no op, assume ifdef format

	script[stack].index = 1;
	if(*v1)
		return true;

	return false;
}		
		
bool ScriptInterp::Push(void)
{
	if(stack >= (SCRIPT_STACK_SIZE - 1))
	{
		Error("stack-overflow");
		return false;
	}

	script[stack + 1] = script[stack];
	++stack;
	return true;
}

bool ScriptInterp::Pull(void)
{
	if(!stack)
	{
		Error("stack-underflow");
		return false;
	}

	if(script[stack - 1].local != script[stack].local)
	{
		if(script[stack].local)
			delete script[stack].local;
	}

	--stack;
	return true;
}

bool ScriptInterp::getOnce(void)
{
	bool ret = once;
	once = false;
	return ret;
}

bool ScriptInterp::Redirect(const char *scriptname)
{
	scriptname_t *scr;
	char namebuf[128];
	char *ext;

	if(!strncmp(scriptname, "::", 2))
	{
		strcpy(namebuf, script[stack].script->name);
		ext = strstr(namebuf, "::");
		if(ext) 
			strcpy(ext, scriptname);
		else
			strcat(namebuf, scriptname);
	}
	else
		strcpy(namebuf, scriptname);
	
	scr = getScriptImage(namebuf);
	if(scr)
	{
		script[stack].script = scr;
		script[stack].line = scr->first;
		return true;
	}
	return false;
}

bool ScriptInterp::Attach(const char *scriptname)
{
	ScriptModule *mod;
	session = NULL;
	stack = 0;
	cmd->EnterMutex();
	image = cmd->active;
	if(!image)
	{
		cmd->LeaveMutex();
		return false;
	}
	setSymbol("script.error", symsize);
	setSymbol("script.trigger", symsize);
	setSymbol("script.home", symsize);
	setSymbol("script.error", "none");
	setSymbol("script.trigger", "none");
	setSymbol("script.home", scriptname);
	script[stack].local = NULL;
	script[stack].script = getScriptImage(scriptname);
	if(script[stack].script)
	{
		mod = ScriptModule::first;
		while(mod)
		{
			mod->moduleAttach(this);
			mod = mod->next;
		}
		script[stack].line = script[stack].script->first;
		script[stack].index = 0;
		script[stack].read = NULL;
		++image->refcount;
		cmd->LeaveMutex();
		return true;
	}
	cmd->LeaveMutex();
	packtoken = ',';
	once = true;
	signalmask = 0;
	logerror(scriptname, getId(), "missing; attach failed");
	return false;	
}

void ScriptInterp::Detach(void)
{
	ScriptModule *mod;

	if(!image)
		return;

	if(session)
	{
		delete session;
		session = NULL;
	}

	cmd->EnterMutex();
	--image->refcount;
	mod = ScriptModule::first;
	while(mod)
	{
		mod->moduleDetach(this);
		mod = mod->next;
	}
	if(image)
		if(!image->refcount && image != cmd->active)
			delete image;

	cmd->LeaveMutex();
	image = NULL;

	while(stack)
		Pull();
	locks.Release(this);
}

int ScriptInterp::initKeywords(int size)
{
	int idx = 0;
	int count = 0;
	scriptline_t *line = script[stack].line;
	char *opt;

	if(!size)
		size = getSymbolSize();

	while(idx < line->argc)
	{
		opt = line->args[idx++];
		if(*opt != '=')
			continue;
		if(*(++opt) == '%')
			++opt;
		++count;
		setSymbol(opt, getSymbolSize());
		setSymbol(opt, line->args[idx++]);
	}
	return count;
}

char *ScriptInterp::getKeyword(const char *kw)
{
	int idx = 0;
	scriptline_t *line = script[stack].line;
	char *opt;
	while(idx < line->argc)
	{
		opt = line->args[idx++];
		if(*opt == '=')
		{
			if(!strnicmp(kw, opt + 1, strlen(kw)))
				return getContent(line->args[idx]);
			++idx;
		}
	}
	return NULL;
}

char *ScriptInterp::getOption(const char *def)
{
	for(;;)
	{
		if(script[stack].index >= script[stack].line->argc)
			return (char *)def;

		if(*script[stack].line->args[script[stack].index] != '=')
			break;

		script[stack].index += 2;
	}		

	return script[stack].line->args[script[stack].index++];
}

scriptsymbol_t *ScriptInterp::getVariable(size_t size)
{
	char *opt;
	scriptsymbol_t *sym;

	for(;;)
	{
		if(script[stack].index >= script[stack].line->argc)
			return NULL;

		opt = script[stack].line->args[script[stack].index++];

		if(*opt != '=')
			break;

		script[stack].index += 2;
	}

	if(*opt != '%' && *opt != '@')
		return NULL;	
	
	if(*opt == '@')
	{
		sym = getIndirect(++opt);
		if(sym)
			return sym;

		if(strchr(opt, '.'))
			sym = getEntry(opt);
		else
			sym = getLocal(opt);

		if(sym)
			opt = sym->data;
		else
			return NULL;
	}
	else
		++opt;

	if(strchr(opt, '.'))
		return getEntry(opt, size);
	else
		return getLocal(opt, size);
}

char *ScriptInterp::getContent(char *opt)
{
	ScriptProperty *prop;
	scriptsymbol_t *sym;
	char *ext = strrchr(opt, '.');
	char *buf;

	if(!opt)
		return NULL;

	if(*opt != '%' && *opt != '@')
		return opt;

	if(*opt == '@')
	{
		++opt;
		sym = getIndirect(opt);
		if(sym)
			return readSymbol(sym);

		sym = getLocal(opt);
		if(!sym)
			return NULL;

		opt = sym->data;
	}
	else
		++opt;

	sym = getLocal(opt);
	if(sym)
		return readSymbol(sym);

	if(!ext)
		return NULL;

	buf = new char[strlen(opt) + 1];
	strcpy(buf, opt);
	*strrchr(buf, '.') = 0;	

	sym = getLocal(buf);
	delete[] buf;
	if(!sym)
		return NULL;

	if(++tempidx >= SCRIPT_TEMP_SPACE)
		tempidx = 0;
	buf = temps[tempidx];

	if(!strnicmp(ext, ".len", 4))
		sprintf(buf, "%d", strlen(sym->data));
	else if(!stricmp(ext, ".size"))
		sprintf(buf, "%d", sym->flags.size);
	else if(!strnicmp(ext, ".val", 4))
		sprintf(buf, "%d", atoi(sym->data));
	else if(!stricmp(ext, ".count"))
		switch(sym->flags.type)
		{
		case SYM_TYPE_SEQUENCE:
		case SYM_TYPE_FIFO:
		case SYM_TYPE_STACK:
			sprintf(buf, "%d", sym->data[3]);
			break;
		default:
			return "1";
		}
	else if(NULL != (prop = ScriptProperty::find(ext + 1)))
		prop->getProperty(sym->data, buf, getSymbolSize());
	else if(!stricmp(ext, ".max"))
		switch(sym->flags.type)
		{
		case SYM_TYPE_SEQUENCE:
		case SYM_TYPE_FIFO:
		case SYM_TYPE_STACK:
			sprintf(buf, "%d", sym->data[4]);
			break;
		default:
			sprintf(buf, "%d", sym->flags.size);
		}
	else if(!stricmp(ext, ".type"))
		switch(sym->flags.type)
		{
		case SYM_TYPE_ALIAS:
			return "alias";
		case SYM_TYPE_INDEX:
			return "index";
		case SYM_TYPE_SEQUENCE:
			return "sequence";
		case SYM_TYPE_STACK:
			return "stack";
		case SYM_TYPE_COUNTER:
			return "counter"; 
		case SYM_TYPE_FIFO:
			return "fifo";
		default:
			return "string";
		}
	else
		buf = NULL;

	return buf;
}

char *ScriptInterp::getValue(const char *def)
{
	char *opt = getOption(NULL);

	if(!opt)
		return (char *)def;

	opt = getContent(opt);
	if(!opt)
		return (char *)def;

	return opt;
}	

const char *ScriptInterp::getMember(void)
{
	char *cmd = script[stack].line->cmd;

	while(*cmd && *cmd != '.')
		++cmd;

	if(*cmd)
		return ++cmd;
	return NULL;
}

